/* 
 * Copyright (c) 2007 Barath Raghavan.
 * Copyright (c) 2007 The Regents of the University of California.
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University of California nor the names of its 
 *    contributors may be used to endorse or promote products derived from 
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <arpa/inet.h>

#ifndef WORDS_BIGENDIAN
#include <byteswap.h>
#endif

#include <stdint.h>
#include <openssl/aes.h>
#include <openssl/rand.h>

#define KEYLEN AES_BLOCK_SIZE // bytes
#define NUM_MASKED_KEY_BYTES 2

/* Does a 64-bit equivalent of ntohl */
uint64_t ntohll(uint64_t netlong) {
#ifndef WORDS_BIGENDIAN
  return bswap_64(netlong);
#else
  return netlong;
#endif
}

/* Does a 64-bit equivalent of htonl */
uint64_t htonll(uint64_t hostlong) {
  return ntohll(hostlong);
}

void xor_blocks(unsigned char *out, unsigned char *in1, unsigned char *in2) {
  for (int i = 0; i < AES_BLOCK_SIZE; i++) {
    out[i] = in1[i] ^ in2[i];
  }
}

void print_block(unsigned char *b) {
  unsigned int *b2 = (unsigned int *) b;
  fprintf(stderr, "%x%x%x%x\n", b2[0], b2[1], b2[2], b2[3]);
}

int main(int argc, char *argv[]) {
  if (argc != 2) {
    printf("Usage: %s FILE > outdata\n", argv[0]);
    exit(1);
  }

  char *filename = argv[1];

  FILE *infile;
  if ((infile = fopen(filename, "rb")) == NULL) {
    perror("fopen");
    exit(1);
  }

  if (fseek(infile, -(sizeof(uint64_t) + AES_BLOCK_SIZE + sizeof(uint32_t)), SEEK_END) < 0) {
    perror("fseek");
    exit(1);
  }

  /* read hash key block */
  unsigned char hashblock[AES_BLOCK_SIZE];
  if (fread(hashblock, AES_BLOCK_SIZE, 1, infile) != 1) {
    perror("fread");
    exit(1);
  }

  /* read in totallen */
  uint64_t totallen;
  if (fread(&totallen, sizeof(uint64_t), 1, infile) != 1) {
    perror("fread");
    exit(1);
  }

  totallen = ntohll(totallen);

  /* read in nummasked */
  uint32_t nn;
  if (fread(&nn, sizeof(uint32_t), 1, infile) != 1) {
    perror("fread");
    exit(1);
  }
  uint32_t nummasked = ntohl(nn);
  if (nummasked != 0 && nummasked != 2 && nummasked != 4) {
    fprintf(stderr, "Error: number of masked keybytes must be 0, 2, or 4.\n");
    exit(1);
  }
  
  if (nummasked != 0) {
    fprintf(stderr, "Brute forcing of key required.\n\nUsing unstaple in this manner on copyrighted material may constitute a violation of the USA Digital Millennium Copyright Act (DMCA) due to the circumvention of a copyright protection mechanism.\n\nDo you wish to proceed [yes/no]? ");
    
    char reply[5];
    fgets(reply, 4, stdin);
    if (strcasecmp(reply, "yes") != 0) {
      fprintf(stderr, "Aborting.\n");
      exit(1);
    }
  }

  rewind(infile);

  /* read in zero ciphertext */
  unsigned char zeroctext[AES_BLOCK_SIZE];
  if (fread(zeroctext, AES_BLOCK_SIZE, 1, infile) != 1) {
    perror("fread");
    exit(1);
  }

  /* compute hashes */
  unsigned char zeroin[AES_BLOCK_SIZE];
  memset(zeroin, 0, AES_BLOCK_SIZE);
  
  AES_KEY zerokey;
  AES_set_encrypt_key(zeroin, AES_BLOCK_SIZE * 8, &zerokey);

  /* file format: E_k(0), AONT of indata, s', totallen */

  /* begin hashing */
  int numblocks = (totallen / AES_BLOCK_SIZE) + ((totallen % AES_BLOCK_SIZE) ? 1 : 0);

  unsigned char ctrptext[AES_BLOCK_SIZE];
  memset(ctrptext, 0, AES_BLOCK_SIZE);

  for (uint64_t i = 1; i <= numblocks; i++) {
    unsigned char ctext[AES_BLOCK_SIZE];
    if (fread(ctext, AES_BLOCK_SIZE, 1, infile) != 1) {
      perror("fread");
      exit(1);
    }

    uint64_t nctr = htonll(i);
    memcpy(ctrptext, &nctr, sizeof(nctr));

    /* hash */
    unsigned char hashctext[AES_BLOCK_SIZE];
    
    xor_blocks(ctext, ctext, ctrptext);
    AES_encrypt(ctext, hashctext, &zerokey);

    xor_blocks(hashblock, hashblock, hashctext);
  }

  /* whatever we have in hashblock at this point should be the truncated key */
  /* begin brute force key finding */
  AES_KEY key;
  AES_set_encrypt_key(hashblock, AES_BLOCK_SIZE * 8, &key);

  unsigned char zeroctext2[AES_BLOCK_SIZE];

  if (nummasked != 0) {
    uint32_t keyprefix = 0;
    uint32_t nkeyprefix;
    int foundkey = 0;

    for (int64_t i = 0; i < ((int64_t) 1 << (int64_t) (8*nummasked)); i++) {
      nkeyprefix = htonl(keyprefix);
      nkeyprefix = bswap_32(nkeyprefix);

      memcpy(hashblock, &nkeyprefix, nummasked);
      AES_set_encrypt_key(hashblock, AES_BLOCK_SIZE * 8, &key);
      
      /* try to encrypt zero with the newly guessed key */
      AES_encrypt(zeroin, zeroctext2, &key);
      
      /* check the encryption */
      if (memcmp(zeroctext, zeroctext2, AES_BLOCK_SIZE) == 0) {
	foundkey = 1;
	break;
      }
      
      keyprefix++;
    }
    
    if (!foundkey) {
      fprintf(stderr, "Couldn't decrypt.  Unstapling failed.\n");
      exit(1);
    }
  } else {
    /* try to encrypt zero with the newly guessed key */
    AES_encrypt(zeroin, zeroctext2, &key);
    
    /* check the encryption */
    if (memcmp(zeroctext, zeroctext2, AES_BLOCK_SIZE) != 0) {
      fprintf(stderr, "Couldn't decrypt.  Unstapling failed.\n");
      exit(1);
    }
  }

  
  /* decrypt data */
  if (fseek(infile, AES_BLOCK_SIZE, SEEK_SET) < 0) {
    perror("fseek");
    exit(1);
  }

  unsigned char ctrctext[AES_BLOCK_SIZE];
  memset(ctrptext, 0, AES_BLOCK_SIZE);

  for (uint64_t i = 1; i <= numblocks; i++) {
    unsigned char ctext[AES_BLOCK_SIZE];
    unsigned char ptext[AES_BLOCK_SIZE];

    if (fread(ctext, AES_BLOCK_SIZE, 1, infile) != 1) {
      perror("fread");
      exit(1);
    }

    uint64_t nctr = htonll(i);
    memcpy(ctrptext, &nctr, sizeof(nctr));

    /* encrypt counter */
    AES_encrypt(ctrptext, ctrctext, &key);

    /* xor with plaintext */
    xor_blocks(ptext, ctext, ctrctext);

    /* output */
    if (i < numblocks || ((totallen % AES_BLOCK_SIZE) == 0)) {
      if (fwrite(ptext, AES_BLOCK_SIZE, 1, stdout) != 1) {
	perror("fwrite");
	exit(1);
      }
    } else {
      if (fwrite(ptext, totallen % AES_BLOCK_SIZE, 1, stdout) != 1) {
	perror("fwrite");
	exit(1);
      }
    }
  }

  fflush(stdout);

  return 0;
}
